% LOCAL LEARNING OF ACTIVE BASIS
%% Load in exponential model, mex C codes, and set parameters
% ExponentialModel; % this line is not needed if 'storedExponentialModel' exists
clear; close all; 
load 'storedExponentialModel'; % load in exponential model 
mex Clearn.c;  % learning from aligned images 
mex CsigmoidTransform.c; % sigmoid transformation 
epsilon = .1; % allowed correlation between selected Gabors 
subsample = 1; % subsample in computing MAX1 and SUM2 maps   
locationShiftLimit = 2; % shift in normal direction = locationShiftLimit*subsample pixels
orientShiftLimit = 1; % shift in orientation
Correlation = CorrFilter(allFilter, epsilon); % correlation between filters 
numNeighbor = 5; % number of nearest neighbors including self
numElement = 40; % number of Gabors in active basis   
numIteration = 3;  % number of iterations
selfWeight = numNeighbor/3.; % weight of self (starting image of local learning)
%% Load in training images
useMinimumSize = -1; resizeFactor = 1.; % resize the input image 
useFixedSize = 1; sizex = 100; sizey = 100; % forced size of inpu image
imageFolder = 'positiveImage'; % folder of training images  
imageName = dir([imageFolder '\*.jpg']);
numImage = size(imageName, 1); % number of training images 
I = cell(1, numImage); 
for (img = 1 : numImage)
    tmpIm = imread([imageFolder '\' imageName(img).name]); 
    if size(tmpIm,3) == 3
        tmpIm = rgb2gray(tmpIm);
    end
    if (useMinimumSize>0)
       I{img} = imresize(single(tmpIm), resizeFactor, 'nearest');
    end
    if (useFixedSize>0)
       I{img} = imresize(single(tmpIm), [sizex sizey], 'nearest');      
    end     
end
if (useMinimumSize>0)   
    imageSize = zeros(numImage, 2);
    for (img = 1 : numImage)
      imageSize(img, :) = size(I{img});   
    end
    sizex = min(imageSize(:, 1)); sizey = min(imageSize(:, 2)); 
    for (img = 1 : numImage)
      I{img} = I{img}(1:sizex, 1:sizey); 
    end
end        
allSizex = zeros(1, numImage) + sizex; 
allSizey = zeros(1, numImage) + sizey; 
disp(['number of training images = ' num2str(numImage)]);
%% Compute SUM1 maps by Gabor filtering
disp(['start filtering']); tic
SUM1map = ApplyFilterfftSame(I, allFilter); % SUM1 maps by Gabor filtering
disp(['filtering time: ' num2str(toc) ' seconds']);
CsigmoidTransform(numImage, allSizex, allSizey, numOrient, saturation, SUM1map);
SUM1mapBackup = cell(numImage, numOrient); 
for (img = 1:numImage)
    for (orient = 1:numOrient)
        SUM1mapBackup{img, orient} = SUM1map{img, orient}+0.; 
    end
end
%% Prepare output variables for learning
selectedOrient = zeros(1, numElement);  % orientation and location of selected Gabors
selectedx = zeros(1, numElement); 
selectedy = zeros(1, numElement); 
selectedlambda = zeros(1, numElement); % weighting parameter for scoring template matching
selectedLogZ = zeros(1, numElement); % normalizing constant
commonTemplate = single(zeros(sizex, sizey)); % template of active basis 
deformedTemplate = cell(1, numImage); % templates for training images 
for (img = 1 : numImage)
    deformedTemplate{img} = single(zeros(sizex, sizey));  
end
SUM2score = zeros(numImage, 1); % template matching scores for training images 
allSUM2score = zeros(numImage, numImage); % matching score starting from each image
meanSUM2score = zeros(numImage, 1); % average matching score starting from each image
neighbor = zeros(numImage, numNeighbor); % neighbors of each starting image
weight = zeros(1, numImage); 
%% start local learning from every image 
for (this = 1:numImage)
disp(['get example ' num2str(this)]);
for (img = 1 : numImage)
    weight(img) = 0.; 
end
weight(this) = 1; % start from single image learning
%% Iteration part 1: learning template from current neighbors
for (it = 1 : numIteration)
%disp(['    estimate the local template in iteration ' num2str(it)]);
tic
if (it==1)
    locationShiftLimitUse = 0; orientShiftLimitUse = 0;  
else
    locationShiftLimitUse = locationShiftLimit; orientShiftLimitUse = orientShiftLimit; 
end
Clearn(numOrient, locationShiftLimitUse, orientShiftLimitUse, subsample, ... % about active basis  
       numElement, numImage, sizex, sizey, SUM1map, SUM1mapBackup, ... % about training images 
       halfFilterSize, Correlation, allSymbol(1, :), ... % about filters
       numStoredPoint, storedlambda, storedExpectation, storedLogZ, ... % about exponential model 
       selectedOrient, selectedx, selectedy, selectedlambda, selectedLogZ, SUM2score, ... % learned parameters
       commonTemplate, deformedTemplate, weight); % learned templates 
%% Iteration part 2: find neighbors using current template
%disp(['    select nearest neighbors in iteration ' num2str(it)]);
[SUM2scoreSort ind] = sort(SUM2score, 'descend'); 
for (img = 1 : numImage)
    weight(img) = 0; 
end
for (k = 1 : numNeighbor)
    weight(ind(k)) = 1.; 
end
weight(this) = selfWeight; 
if (it==numIteration)
    meanSUM2score(this, 1) = mean(SUM2scoreSort(1:numNeighbor)); 
    allSUM2score(this, :) = SUM2score; 
    neighbor(this, :) = ind(1:numNeighbor); 
    allTemplate{this} = commonTemplate+0.; 
    Outthis; 
end
clear SUM2scoreSort; 
clear ind; 
end
save 'templates' allTemplate allSUM2score meanSUM2score sizex sizey numImage;
end
OutAllTemplate; 


